import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from .logging_context import worker_id_context, worker_type_context


@dataclass
class WorkerSettings:
    pass


class Worker(ABC):
    settings: WorkerSettings
    worker_type: str
    worker_id: str

    def __init_subclass__(cls, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)
        if not hasattr(cls, "worker_type"):
            cls.worker_type = f"{cls.__module__}.{cls.__name__}"

    def __init__(self, settings: WorkerSettings | None) -> None:
        self.settings = settings or WorkerSettings()
        self.worker_id = str(uuid.uuid4())[:8]

    def __str__(self) -> str:
        return f"{self.worker_type}-{self.worker_id}"

    def setup_logging_context(self) -> None:
        if self.worker_type:
            worker_type_context.set(self.worker_type)
        if self.worker_id:
            worker_id_context.set(self.worker_id)

    @abstractmethod
    async def run(self, burst: bool = False) -> None:  # pragma: no cover
        pass



import asyncio
from asyncio.exceptions import CancelledError
from dataclasses import dataclass
from datetime import timedelta
from typing import cast
from unittest import mock

import pytest

from aiosafeconsumer import (
    Worker,
    WorkerDef,
    WorkerPool,
    WorkerPoolSettings,
    WorkerSettings,
)
from aiosafeconsumer.run import run as cli_run


@dataclass
class AllWorkerSettings(WorkerSettings):
    counter: dict[str, int]
    start_counter: dict[str, int]
    error_on: int | None = None
    exit_on: int | None = None


class BaseWorker(Worker):
    settings: AllWorkerSettings

    async def run(self, burst: bool = False) -> None:
        counter = self.settings.counter
        counter.setdefault(self.worker_type, 0)

        start_counter = self.settings.start_counter
        start_counter.setdefault(self.worker_type, 0)
        start_counter[self.worker_type] += 1

        while True:
            counter[self.worker_type] += 1
            cnt = counter[self.worker_type]

            if self.settings.error_on and cnt == self.settings.error_on:
                raise Exception("Test error")
            if self.settings.exit_on and cnt == self.settings.exit_on:
                return

            await asyncio.sleep(0.001)
            if burst:
                break


class Worker1(BaseWorker):
    worker_type = "worker1"


class Worker2(BaseWorker):
    worker_type = "worker2"


class Worker3(BaseWorker):
    worker_type = "worker3"


@pytest.fixture
def counter() -> dict[str, int]:
    return {}


@pytest.fixture
def start_counter() -> dict[str, int]:
    return {}


@pytest.fixture
def worker_pool_settings(
    counter: dict[str, int], start_counter: dict[str, int]
) -> WorkerPoolSettings:
    workers = [
        WorkerDef(
            worker_class=Worker1,
            worker_settings=AllWorkerSettings(
                counter=counter,
                start_counter=start_counter,
            ),
        ),
        WorkerDef(
            worker_class=Worker2,
            worker_settings=AllWorkerSettings(
                counter=counter,
                start_counter=start_counter,
            ),
        ),
        WorkerDef(
            worker_class=Worker3,
            worker_settings=AllWorkerSettings(
                counter=counter,
                start_counter=start_counter,
            ),
            worker_group="worker3",
        ),
    ]
    return WorkerPoolSettings(workers=workers)


@pytest.mark.asyncio
async def test_worker_pool_run(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    pool = WorkerPool(worker_pool_settings)

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    task.cancel()

    with pytest.raises(CancelledError):
        await task

    assert counter["worker1"] > 10
    assert counter["worker2"] > 10
    assert counter["worker3"] > 10

    assert start_counter["worker1"] == 1
    assert start_counter["worker2"] == 1
    assert start_counter["worker3"] == 1


@pytest.mark.asyncio
async def test_worker_pool_run_with_concurrency(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    pool = WorkerPool(worker_pool_settings, override_concurrency=4)

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    task.cancel()

    with pytest.raises(CancelledError):
        await task

    assert counter["worker1"] > 10
    assert counter["worker2"] > 10
    assert counter["worker3"] > 10

    assert start_counter["worker1"] == 4
    assert start_counter["worker2"] == 4
    assert start_counter["worker3"] == 4


@pytest.mark.asyncio
async def test_worker_pool_run_with_error(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    worker_settings = cast(
        AllWorkerSettings, worker_pool_settings.workers[0].worker_settings
    )
    worker_settings.error_on = 2
    pool = WorkerPool(worker_pool_settings, sleep_on_task_exit=timedelta(seconds=0.001))

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    task.cancel()

    with pytest.raises(CancelledError):
        await task

    assert counter["worker1"] > 10
    assert counter["worker2"] > 10
    assert counter["worker3"] > 10

    assert start_counter["worker1"] == 2
    assert start_counter["worker2"] == 1
    assert start_counter["worker3"] == 1


@pytest.mark.asyncio
async def test_worker_pool_run_with_exit(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    worker_settings = cast(
        AllWorkerSettings, worker_pool_settings.workers[0].worker_settings
    )
    worker_settings.exit_on = 2
    pool = WorkerPool(worker_pool_settings, sleep_on_task_exit=timedelta(seconds=0.001))

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    task.cancel()

    with pytest.raises(CancelledError):
        await task

    assert counter["worker1"] > 10
    assert counter["worker2"] > 10
    assert counter["worker3"] > 10

    assert start_counter["worker1"] == 2
    assert start_counter["worker2"] == 1
    assert start_counter["worker3"] == 1


@pytest.mark.asyncio
async def test_worker_pool_run_terminate(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    pool = WorkerPool(worker_pool_settings)

    assert counter == {}

    task = asyncio.create_task(pool.run())
    await asyncio.sleep(0.1)
    pool.terminate()
    await task

    assert counter["worker1"] > 10
    assert counter["worker2"] > 10
    assert counter["worker3"] > 10

    assert start_counter["worker1"] == 1
    assert start_counter["worker2"] == 1
    assert start_counter["worker3"] == 1


@pytest.mark.asyncio
async def test_worker_pool_run_burst(
    counter: dict[str, int],
    start_counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    assert counter == {}

    pool = WorkerPool(worker_pool_settings, burst=True)
    await pool.run()

    assert counter == {
        "worker1": 1,
        "worker2": 1,
        "worker3": 1,
    }

    assert start_counter["worker1"] == 1
    assert start_counter["worker2"] == 1
    assert start_counter["worker3"] == 1


def test_cli_list_tasks(worker_pool_settings: WorkerPoolSettings) -> None:
    with mock.patch("aiosafeconsumer.run.print") as print_m:
        cli_run(
            ["test", "--list-tasks"],
            pool_class=WorkerPool,
            pool_settings=worker_pool_settings,
        )

        assert print_m.mock_calls == [
            mock.call("worker1"),
            mock.call("worker2"),
            mock.call("worker3"),
        ]


def test_cli_list_groups(worker_pool_settings: WorkerPoolSettings) -> None:
    with mock.patch("aiosafeconsumer.run.print") as print_m:
        cli_run(
            ["test", "--list-groups"],
            pool_class=WorkerPool,
            pool_settings=worker_pool_settings,
        )

        assert print_m.mock_calls == [
            mock.call("default"),
            mock.call("worker3"),
        ]


def test_cli_run(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    cli_run(
        ["test", "--burst"],
        pool_class=WorkerPool,
        pool_settings=worker_pool_settings,
    )

    assert counter == {
        "worker1": 1,
        "worker2": 1,
        "worker3": 1,
    }


def test_cli_run_task(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    cli_run(
        ["test", "--burst", "--task=worker1"],
        pool_class=WorkerPool,
        pool_settings=worker_pool_settings,
    )

    assert counter == {
        "worker1": 1,
    }


def test_cli_run_exclude_task(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    cli_run(
        ["test", "--burst", "--exclude-task=worker1"],
        pool_class=WorkerPool,
        pool_settings=worker_pool_settings,
    )

    assert counter == {
        "worker2": 1,
        "worker3": 1,
    }


def test_cli_run_group(
    counter: dict[str, int],
    worker_pool_settings: WorkerPoolSettings,
) -> None:
    cli_run(
        ["test", "--burst", "--group=worker3"],
        pool_class=WorkerPool,
        pool_settings=worker_pool_settings,
    )

    assert counter == {
        "worker3": 1,
    }



import pickle
from dataclasses import dataclass
from typing import Any, cast

from aiosafeconsumer.datasync import EnumerateIDsRecord, EventType, ObjectID, Version
from aiosafeconsumer.datasync.redis import RedisWriter, RedisWriterSettings

from ..types import User, UserDeleteRecord, UserEnumerateRecord, UserRecord


@dataclass
class UsersRedisWriterSettings(RedisWriterSettings[User]):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(
            version_getter=self._version_getter,
            event_type_getter=self._event_type_getter,
            id_getter=self._id_getter,
            enum_getter=self._enum_getter,
            version_serializer=self._version_serializer,
            version_deserializer=self._version_deserializer,
            record_serializer=self._record_serializer,
            **kwargs,
        )

    @staticmethod
    def _version_getter(item: User) -> Version:
        return int(item.ev_time.timestamp())

    @staticmethod
    def _record_serializer(item: User) -> bytes:
        return pickle.dumps(item)

    @staticmethod
    def _event_type_getter(item: User) -> EventType:
        return item.ev_type

    @staticmethod
    def _id_getter(item: User) -> ObjectID:
        assert isinstance(item, UserRecord) or isinstance(item, UserDeleteRecord)
        return item.id

    @staticmethod
    def _enum_getter(item: User) -> EnumerateIDsRecord:
        assert isinstance(item, UserEnumerateRecord)
        return EnumerateIDsRecord(ids=cast(list[ObjectID], item.ids))

    @staticmethod
    def _version_serializer(ver: Version) -> bytes:
        return str(ver).encode()

    @staticmethod
    def _version_deserializer(val: bytes) -> Version:
        return int(val.decode())


class UsersRedisWriter(RedisWriter[User]):
    settings: UsersRedisWriterSettings



import asyncio
import itertools
import logging
import logging.config
import os
from collections.abc import AsyncGenerator
from dataclasses import dataclass

import pytest

from aiosafeconsumer import (
    ConsumerWorker,
    ConsumerWorkerSettings,
    DataProcessor,
    DataProcessorSettings,
    DataSource,
    DataSourceSettings,
    DataTransformer,
    DataTransformerSettings,
)

collect_ignore = []
if not os.getenv("KAFKA_BOOTSTRAP_SERVERS") and not os.getenv("REDIS_URL"):
    collect_ignore.append("real/kafka_redis")
if not os.getenv("KAFKA_BOOTSTRAP_SERVERS") and not os.getenv("ELASTICSEARCH_URL"):
    collect_ignore.append("real/kafka_elasticsearch")


_LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "filters": {
        "context_injecting": {
            "()": "aiosafeconsumer.logging.ContextInjectingFilter",
        },
        "is_worker_context": {
            "()": "aiosafeconsumer.logging.IsWorkerContextFilter",
        },
        "is_not_worker_context": {
            "()": "aiosafeconsumer.logging.IsWorkerContextFilter",
            "invert": True,
        },
    },
    "formatters": {
        "common": {
            "format": "[%(levelname)s/%(name)s] %(message)s",
        },
        "worker": {
            "()": "aiosafeconsumer.logging.ExtraFieldsFormatter",
            "fmt": "[%(levelname)s/%(worker_type)s-%(worker_id)s] %(message)s",
        },
    },
    "handlers": {
        "common": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "common",
            "filters": ["is_not_worker_context"],
        },
        "worker": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "worker",
            "filters": ["is_worker_context", "context_injecting"],
        },
    },
    "loggers": {
        "aiosafeconsumer": {
            "level": "DEBUG",
            "handlers": ["common", "worker"],
            "propagate": True,
        },
    },
}


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup("aiosafeconsumer")
    group.addoption(
        "--with-custom-logging",
        action="store_true",
        dest="with_custom_logging",
        default=False,
        help="Use ContextInjectingFilter and ExtraFieldsFormatter",
    )


def pytest_configure(config: pytest.Config) -> None:
    with_custom_logging = config.getoption("with_custom_logging")
    if with_custom_logging:
        logging.config.dictConfig(_LOGGING)


class StrSourceSettings(DataSourceSettings):
    stop_on: int | None = None


class StrSource(DataSource[str]):
    settings: StrSourceSettings

    BATCHES = [
        ["one", "two", "three", "fore", "five"],
        ["six", "seven", "eight", "nine", "ten"],
    ]

    is_resource_allocated = False

    async def read(self) -> AsyncGenerator[list[str], None]:
        self.is_resource_allocated = True
        count = 0
        try:
            for batch in itertools.cycle(self.BATCHES):
                yield batch
                count += 1
                if self.settings.stop_on is not None and count == self.settings.stop_on:
                    break
        finally:
            self.is_resource_allocated = False


class StrProcessor(DataProcessor[str]):
    storage: list[str]

    def __init__(self, settings: DataProcessorSettings | None = None) -> None:
        super().__init__(settings)
        self.storage = []

    async def process(self, batch: list[str]) -> None:
        self.storage.extend(batch)
        await asyncio.sleep(0.1)
        if len(self.storage) >= 1000:
            raise Exception("Too many data")


class IntProcessor(DataProcessor[int]):
    storage: list[int]

    def __init__(self, settings: DataProcessorSettings | None = None) -> None:
        super().__init__(settings)
        self.storage = []

    async def process(self, batch: list[int]) -> None:
        self.storage.extend(batch)
        await asyncio.sleep(0.1)
        if len(self.storage) >= 1000:
            raise Exception("Too many data")


@dataclass
class StrToIntTransformerSettings(DataTransformerSettings[str, int]):
    pass


class StrToIntTransformer(DataTransformer[str, int]):
    _MAPPING = {
        "one": 1,
        "two": 2,
        "three": 3,
        "fore": 4,
        "five": 5,
        "six": 6,
        "seven": 7,
        "eight": 8,
        "nine": 9,
        "ten": 10,
    }

    async def transform(self, batch: list[str]) -> list[int]:
        return [self._MAPPING[x] for x in batch]


@dataclass
class StrConsumerSettings(ConsumerWorkerSettings[str]):
    pass


@dataclass
class StrConsumerWithTansformerSettings(ConsumerWorkerSettings[str]):
    pass


@pytest.fixture
def source_settings() -> StrSourceSettings:
    return StrSourceSettings()


@pytest.fixture
def source(source_settings: StrSourceSettings) -> StrSource:
    return StrSource(source_settings)


@pytest.fixture
def consumer_settings(source_settings: StrSourceSettings) -> StrConsumerSettings:
    return StrConsumerSettings(
        source_class=StrSource,
        source_settings=source_settings,
        processor_class=StrProcessor,
        processor_settings=DataProcessorSettings(),
    )


@pytest.fixture
def transformer_settings() -> StrToIntTransformerSettings:
    return StrToIntTransformerSettings(
        target_processor_class=IntProcessor,
        target_processor_settings=DataProcessorSettings(),
    )


@pytest.fixture
def consumer_with_transformer_settings(
    source_settings: StrSourceSettings,
    transformer_settings: StrToIntTransformerSettings,
) -> StrConsumerWithTansformerSettings:
    return StrConsumerWithTansformerSettings(
        source_class=StrSource,
        source_settings=source_settings,
        processor_class=StrToIntTransformer,
        processor_settings=transformer_settings,
    )


@pytest.fixture
def processor() -> StrProcessor:
    return StrProcessor()


@pytest.fixture
def transformer(
    transformer_settings: StrToIntTransformerSettings,
) -> StrToIntTransformer:
    return StrToIntTransformer(transformer_settings)


@pytest.fixture
def consumer(consumer_settings: StrConsumerSettings) -> ConsumerWorker:
    consumer = ConsumerWorker(consumer_settings)
    return consumer


@pytest.fixture
def consumer_with_transformer(
    consumer_with_transformer_settings: StrConsumerWithTansformerSettings,
) -> ConsumerWorker:
    consumer = ConsumerWorker(consumer_with_transformer_settings)
    return consumer



from dataclasses import dataclass

from aiosafeconsumer import ConsumerWorker, ConsumerWorkerSettings

from .types import User


@dataclass
class UsersWorkerSettings(ConsumerWorkerSettings[User]):
    pass


class UsersWorker(ConsumerWorker):
    worker_type = "sync_users"



from aiosafeconsumer.kafka import KafkaSource, KafkaSourceSettings

from .types import User


class UsersKafkaSourceSettings(KafkaSourceSettings[User]):
    pass


class UsersKafkaSource(KafkaSource[User]):
    settings: UsersKafkaSourceSettings



import os
from collections.abc import AsyncGenerator
from datetime import timedelta

import pytest
import pytest_asyncio
from redis.asyncio import ConnectionPool, Redis

from aiosafeconsumer import WorkerDef, WorkerPoolSettings

from ..deserializers import json_to_namedtuple_deserializer
from ..sources import UsersKafkaSource, UsersKafkaSourceSettings
from ..types import UserDeleteRecord, UserEnumerateRecord, UserEOSRecord, UserRecord
from ..workers import UsersWorker, UsersWorkerSettings
from .processors import UsersRedisWriter, UsersRedisWriterSettings


@pytest.fixture
def redis_url() -> str:
    return os.getenv("REDIS_URL", "")


@pytest_asyncio.fixture
async def redis_pool(redis_url: str) -> AsyncGenerator[ConnectionPool]:
    pool: ConnectionPool = ConnectionPool.from_url(redis_url)
    try:
        yield pool
    finally:
        redis = Redis(connection_pool=pool)
        await redis.flushdb()
        await pool.disconnect()


@pytest.fixture
def redis(redis_pool: ConnectionPool) -> Redis:
    return Redis(connection_pool=redis_pool)


@pytest.fixture
def worker_pool_settings(
    redis: Redis,
    kafka_bootstrap_servers: list[str],
) -> WorkerPoolSettings:
    pool_settings = WorkerPoolSettings(
        workers=[
            WorkerDef(
                worker_class=UsersWorker,
                worker_settings=UsersWorkerSettings(
                    source_class=UsersKafkaSource,
                    source_settings=UsersKafkaSourceSettings(
                        topics=["users"],
                        bootstrap_servers=kafka_bootstrap_servers,
                        group_id="sync_users",
                        value_deserializer=json_to_namedtuple_deserializer(
                            UserRecord,
                            UserDeleteRecord,
                            UserEnumerateRecord,
                            UserEOSRecord,
                        ),
                        getmany_timeout=timedelta(seconds=0.1),
                        kwargs={
                            "auto_offset_reset": "earliest",
                            "fetch_max_wait_ms": 100,
                        },
                    ),
                    processor_class=UsersRedisWriter,
                    processor_settings=UsersRedisWriterSettings(
                        redis=lambda: redis,
                        key_prefix="user:",
                        versions_key="users",
                    ),
                ),
            ),
        ],
    )
    return pool_settings



from collections.abc import AsyncGenerator, Generator
from dataclasses import dataclass
from typing import Any, NamedTuple, cast
from unittest import mock

import pytest
from elasticsearch import AsyncElasticsearch

from aiosafeconsumer.datasync import (
    EnumerateIDsChunk,
    EnumerateIDsRecord,
    EventType,
    ObjectID,
    Version,
)
from aiosafeconsumer.datasync.elasticsearch import (
    Action,
    Document,
    ElasticsearchWriter,
    ElasticsearchWriterSettings,
)


class ItemRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int
    data: str


class ItemDeleteRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int


class ItemEnumerateRecord(NamedTuple):
    ev_type: EventType
    version: int
    ids: list[int]
    chunk_session: str | None = None
    chunk_index: int | None = None
    total_chunks: int | None = None


class ItemEOSRecord(NamedTuple):
    ev_type: EventType
    version: int


Item = ItemRecord | ItemDeleteRecord | ItemEnumerateRecord | ItemEOSRecord


@dataclass
class Settings(ElasticsearchWriterSettings[Item]):
    pass


class Writer(ElasticsearchWriter[Item]):
    settings: Settings


@pytest.fixture
def elasticsearch_mock() -> Generator[AsyncElasticsearch]:
    documents: dict[str, Document] = {}
    versions: dict[str, int] = {}

    async def async_bulk(
        es: AsyncElasticsearch,
        actions: AsyncGenerator[Action],
        pipeline: str | None,
        raise_on_error: bool,
    ) -> None:
        async for act in actions:
            cur_ver = versions.get(act["_id"], -1)
            if cur_ver >= act["_version"]:
                continue

            if act["_op_type"] == "delete":
                documents.pop(act["_id"], None)
                versions.pop(act["_id"], None)
            elif act["_op_type"] == "index":
                documents[act["_id"]] = act["_source"]
                versions[act["_id"]] = act["_version"]

    async def delete_by_query(
        index: str,
        query: dict[str, Any],
        conflicts: str,
        refresh: bool,
        wait_for_completion: bool,
    ) -> None:
        filter_ = query.get("bool", {}).get("filter", {})
        if len(filter_) == 2:
            f1, f2 = filter_
            enum_ids = set(
                str(x) for x in f1.get("bool", {}).get("must_not", {}).get("ids", [])
            )
            enum_ver = f2.get("range", {}).get("ev_time", {}).get("lt")
            if enum_ids and enum_ver:
                for _id, ver in list(versions.items()):
                    if _id not in enum_ids and ver < enum_ver:
                        documents.pop(_id, None)
                        versions.pop(_id, None)
        elif len(filter_) == 1:
            (f1,) = filter_
            eos_ver = f1.get("range", {}).get("ev_time", {}).get("lt")
            if eos_ver:
                for _id, ver in list(versions.items()):
                    if ver < eos_ver:
                        documents.pop(_id, None)
                        versions.pop(_id, None)

    es_m = mock.MagicMock(spec=AsyncElasticsearch)
    es_m.delete_by_query.side_effect = delete_by_query

    es_m._test_documents = documents
    es_m._test_versions = versions

    with mock.patch(
        "aiosafeconsumer.datasync.elasticsearch.async_bulk"
    ) as async_bulk_m:
        async_bulk_m.side_effect = async_bulk
        yield es_m


@pytest.fixture
def settings(elasticsearch_mock: AsyncElasticsearch) -> Settings:
    def get_elasticsearch() -> AsyncElasticsearch:
        return elasticsearch_mock

    def version_getter(item: Item) -> Version:
        return item.version

    def record_serializer(item: Item) -> Document:
        return item._asdict()

    def event_type_getter(item: Item) -> EventType:
        return item.ev_type

    def id_getter(item: Item) -> ObjectID:
        assert isinstance(item, ItemRecord) or isinstance(item, ItemDeleteRecord)
        return item.id

    def enum_getter(item: Item) -> EnumerateIDsRecord:
        assert isinstance(item, ItemEnumerateRecord)
        ids = cast(list[ObjectID], item.ids)
        chunk: EnumerateIDsChunk | None = None
        if item.chunk_session:
            assert item.chunk_index is not None
            assert item.total_chunks is not None
            chunk = EnumerateIDsChunk(
                chunk_index=item.chunk_index,
                total_chunks=item.total_chunks,
                session=item.chunk_session,
            )
        return EnumerateIDsRecord(ids=ids, chunk=chunk)

    settings = Settings(
        elasticsearch=get_elasticsearch,
        index="items",
        version_field="ev_time",
        version_getter=version_getter,
        record_serializer=record_serializer,
        event_type_getter=event_type_getter,
        id_getter=id_getter,
        enum_getter=enum_getter,
    )
    return settings


@pytest.mark.asyncio
async def test_redis_writer_initial_empty(
    elasticsearch_mock: AsyncElasticsearch, settings: Settings
) -> None:
    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=1, id=3, data="3v1"),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert elasticsearch_mock._test_versions == {  # type: ignore
        "1": 1,
        "2": 2,
        "3": 1,
    }

    assert elasticsearch_mock._test_documents == {  # type: ignore
        "1": {"ev_type": "create", "version": 1, "id": 1, "data": "1v1"},
        "2": {"ev_type": "update", "version": 2, "id": 2, "data": "2v2"},
        "3": {"ev_type": "refresh", "version": 1, "id": 3, "data": "3v1"},
    }


@pytest.mark.asyncio
async def test_redis_writer_upsert(
    elasticsearch_mock: AsyncElasticsearch, settings: Settings
) -> None:
    elasticsearch_mock._test_versions.update(  # type: ignore
        {
            "2": 3,
            "3": 1,
            "4": 0,
            "5": 2,
        },
    )
    elasticsearch_mock._test_documents.update(  # type:ignore
        {
            "2": {},
            "3": {},
            "4": {},
            "5": {},
        },
    )

    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=2, id=3, data="3v2"),
        ItemDeleteRecord(ev_type=EventType.DELETE, version=1, id=4),
        ItemDeleteRecord(ev_type=EventType.DELETE, version=1, id=5),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert elasticsearch_mock._test_versions == {  # type: ignore
        "1": 1,
        "2": 3,
        "3": 2,
        "5": 2,
    }
    assert elasticsearch_mock._test_documents == {  # type: ignore
        "1": {"ev_type": "create", "version": 1, "id": 1, "data": "1v1"},
        "2": {},
        "3": {"ev_type": "refresh", "version": 2, "id": 3, "data": "3v2"},
        "5": {},
    }


@pytest.mark.asyncio
async def test_redis_writer_enumerate(
    elasticsearch_mock: AsyncElasticsearch, settings: Settings
) -> None:
    elasticsearch_mock._test_versions.update(  # type: ignore
        {
            "1": 0,
            "2": 2,
            "3": 1,
            "4": 0,
        },
    )
    elasticsearch_mock._test_documents.update(  # type:ignore
        {
            "1": {},
            "2": {},
            "3": {},
            "4": {},
        },
    )

    items: list[Item] = [
        ItemEnumerateRecord(ev_type=EventType.ENUMERATE, version=1, ids=[1, 3]),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert elasticsearch_mock._test_versions == {  # type: ignore
        "1": 0,
        "2": 2,
        "3": 1,
    }
    assert elasticsearch_mock._test_documents == {  # type: ignore
        "1": {},
        "2": {},
        "3": {},
    }


@pytest.mark.asyncio
async def test_redis_writer_eos(
    elasticsearch_mock: AsyncElasticsearch, settings: Settings
) -> None:
    settings.process_eos = True

    elasticsearch_mock._test_versions.update(  # type: ignore
        {
            "1": 0,
            "2": 2,
            "3": 1,
            "4": 0,
        },
    )
    elasticsearch_mock._test_documents.update(  # type:ignore
        {
            "1": {},
            "2": {},
            "3": {},
            "4": {},
        },
    )

    items: list[Item] = [
        ItemEOSRecord(ev_type=EventType.EOS, version=1),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert elasticsearch_mock._test_versions == {  # type: ignore
        "2": 2,
        "3": 1,
    }
    assert elasticsearch_mock._test_documents == {  # type: ignore
        "2": {},
        "3": {},
    }



import asyncio
import logging
import time
from dataclasses import dataclass
from typing import Generic

from .processor import DataProcessor, DataProcessorSettings
from .source import DataSource, DataSourceSettings
from .types import DataType
from .worker import Worker, WorkerSettings

log = logging.getLogger(__name__)


@dataclass
class ConsumerWorkerSettings(Generic[DataType], WorkerSettings):
    source_class: type[DataSource[DataType]]
    source_settings: DataSourceSettings
    processor_class: type[DataProcessor[DataType]]
    processor_settings: DataProcessorSettings


class ConsumerWorker(Worker):
    settings: ConsumerWorkerSettings

    def __init__(self, settings: ConsumerWorkerSettings):
        super().__init__(settings)
        self.source = self.settings.source_class(self.settings.source_settings)
        self.processor = self.settings.processor_class(self.settings.processor_settings)

    async def run(self, burst: bool = False) -> None:
        self.setup_logging_context()

        log.debug(f"Enter {self}")
        generator = self.source.read()
        try:
            while True:
                try:
                    batch = await anext(generator)
                except StopAsyncIteration:
                    log.debug(f"Source {self.source} generator exist")
                    break
                log.info(f"Got batch of {len(batch)} items from {self.source}")

                start_time = time.monotonic()
                await self.processor.process(batch)
                total_time = time.monotonic() - start_time

                log.info(
                    f"{len(batch)} items was processed by {self.processor} in "
                    f"{total_time:.03f} seconds",
                    extra={
                        "items_processed": len(batch),
                        "processing_time": total_time,
                    },
                )

                await self.source.commit(batch)
                await asyncio.sleep(0)

                if burst:
                    log.debug("Performed one cycle in burst mode")
                    break
        finally:
            log.debug(f"Exit {self}")
            await generator.aclose()



import asyncio
import pickle
from datetime import datetime, timedelta, timezone
from decimal import Decimal
from typing import cast

import pytest
from aiokafka import AIOKafkaProducer  # type: ignore
from redis.asyncio import Redis

from aiosafeconsumer import WorkerPool, WorkerPoolSettings
from aiosafeconsumer.datasync import EventType

from ..types import UserRecord


@pytest.fixture
def ev_time() -> datetime:
    return datetime.now(timezone.utc).replace(microsecond=0)


@pytest.fixture
def users(ev_time: datetime) -> list[UserRecord]:
    ev_type = EventType.REFRESH
    ev_source = "test"
    return [
        UserRecord(
            ev_time=ev_time,
            ev_type=ev_type,
            ev_source=ev_source,
            id=1,
            email="user1@example.com",
            score=Decimal("1.23"),
            is_active=True,
        ),
        UserRecord(
            ev_time=ev_time,
            ev_type=ev_type,
            ev_source=ev_source,
            id=2,
            email="user2@example.com",
            score=Decimal("4.56"),
            is_active=False,
        ),
    ]


def encode_version(dt: datetime) -> bytes:
    return str(int(dt.timestamp())).encode()


def encode_id(id: int) -> bytes:
    return str(id).encode()


@pytest.mark.asyncio
async def test_on_empty_db(
    worker_pool_settings: WorkerPoolSettings,
    producer: AIOKafkaProducer,
    users: list[UserRecord],
    redis: Redis,
    ev_time: datetime,
) -> None:

    pool = WorkerPool(worker_pool_settings, burst=True)
    task = asyncio.create_task(pool.run())

    await asyncio.sleep(0.1)
    await producer.start()
    try:
        for user in users:
            await producer.send("users", user)
    finally:
        await producer.flush()
        await producer.stop()

    await task

    versions = await redis.hgetall(b"users")
    assert versions == {
        encode_id(1): encode_version(ev_time),
        encode_id(2): encode_version(ev_time),
    }

    users_in_redis = [
        pickle.loads(cast(bytes, await redis.get(b"user:" + user_id)))
        for user_id in versions.keys()
    ]

    assert users_in_redis == users


@pytest.mark.asyncio
async def test_update(
    worker_pool_settings: WorkerPoolSettings,
    producer: AIOKafkaProducer,
    users: list[UserRecord],
    redis: Redis,
    ev_time: datetime,
) -> None:
    initial_users: list[UserRecord] = []
    for record in users:
        if record.id == 1:
            record = record._replace(ev_time=ev_time + timedelta(minutes=1))
        elif record.id == 2:
            record = record._replace(ev_time=ev_time - timedelta(minutes=1))
        initial_users.append(record)

    versions = {
        encode_id(record.id): encode_version(record.ev_time) for record in initial_users
    }
    await redis.hset(b"users", mapping=versions)  # type: ignore
    for record in initial_users:
        await redis.set(b"user:" + encode_id(record.id), pickle.dumps(record))

    pool = WorkerPool(worker_pool_settings, burst=True)
    task = asyncio.create_task(pool.run())

    await asyncio.sleep(0.1)
    await producer.start()
    try:
        for user in users:
            await producer.send("users", user)
    finally:
        await producer.flush()
        await producer.stop()

    await task

    versions = await redis.hgetall(b"users")
    assert versions == {
        encode_id(1): encode_version(ev_time + timedelta(minutes=1)),
        encode_id(2): encode_version(ev_time),
    }

    users_in_redis = [
        pickle.loads(cast(bytes, await redis.get(b"user:" + user_id)))
        for user_id in versions.keys()
    ]

    assert users_in_redis[0] == initial_users[0]
    assert users_in_redis[1] == users[1]



from datetime import datetime
from decimal import Decimal
from typing import NamedTuple

from aiosafeconsumer.datasync import EventType


class UserRecord(NamedTuple):
    ev_time: datetime
    ev_type: EventType
    ev_source: str
    id: int
    email: str
    score: Decimal
    is_active: bool


class UserDeleteRecord(NamedTuple):
    ev_time: datetime
    ev_type: EventType
    ev_source: str
    id: int


class UserEnumerateRecord(NamedTuple):
    ev_time: datetime
    ev_type: EventType
    ev_source: str
    ids: list[int]
    chunk_session: str | None = None
    chunk_index: int | None = None
    total_chunks: int | None = None


class UserEOSRecord(NamedTuple):
    ev_time: datetime
    ev_type: EventType
    ev_source: str


User = UserRecord | UserDeleteRecord | UserEnumerateRecord | UserEOSRecord



import json
from collections.abc import Sequence
from logging import Filter, Formatter, LogRecord
from typing import Any, cast

from .logging_context import worker_id_context, worker_type_context

CONTEXT_FIELDS = {
    "worker_type": worker_type_context,
    "worker_id": worker_id_context,
}

OTHER_FIELDS = ["items_processed", "processing_time"]


class ContextInjectingFilter(Filter):
    def filter(self, record: LogRecord) -> bool:
        for field, var in CONTEXT_FIELDS.items():
            if not hasattr(record, field):
                value = var.get()
                if value is not None:
                    setattr(record, field, value)
        return True


class IsWorkerContextFilter(Filter):
    def __init__(self, invert: bool = False) -> None:
        super().__init__()
        self.invert = invert

    def filter(self, record: LogRecord) -> bool:
        result = (
            worker_type_context.get() is not None
            and worker_id_context.get() is not None
        )
        if self.invert:
            result = not result
        return result


class ExtraFieldsFormatter(Formatter):
    fields: Sequence[str]
    json_encoder = json.JSONEncoder

    def __init__(self, *args: Any, **kwargs: Any):
        self.fields = cast(
            Sequence[str],
            kwargs.pop("fields", list(CONTEXT_FIELDS.keys()) + OTHER_FIELDS),
        )
        super().__init__(**kwargs)

    def format(self, record: LogRecord) -> str:
        extra = {
            field: getattr(record, field)
            for field in self.fields
            if hasattr(record, field)
        }
        text = super().format(record)
        if extra:
            text += " " + json.dumps(extra, cls=self.json_encoder, ensure_ascii=False)
        return text



from typing import cast

import pytest

from .conftest import IntProcessor, StrProcessor, StrSource, StrToIntTransformer


@pytest.mark.asyncio
async def test_processor(source: StrSource, processor: StrProcessor) -> None:
    assert processor.storage == []

    await processor.process(source.BATCHES[0])
    await processor.process(source.BATCHES[1])

    assert processor.storage == [
        *source.BATCHES[0],
        *source.BATCHES[1],
    ]


@pytest.mark.asyncio
async def test_transformer(source: StrSource, transformer: StrToIntTransformer) -> None:
    target = cast(IntProcessor, transformer.target_processor)

    assert target.storage == []

    await transformer.process(source.BATCHES[0])
    await transformer.process(source.BATCHES[1])

    assert target.storage == list(range(1, 11))



from collections.abc import Callable
from datetime import datetime, timezone
from decimal import Decimal

import pytest

from aiosafeconsumer.datasync import EventType

from .deserializers import json_to_namedtuple_deserializer
from .types import UserDeleteRecord, UserEnumerateRecord, UserEOSRecord, UserRecord


@pytest.fixture()
def deserializer() -> Callable:
    return json_to_namedtuple_deserializer(
        record_class=UserRecord,
        delete_record_class=UserDeleteRecord,
        enumerate_record_class=UserEnumerateRecord,
        eos_record_class=UserEOSRecord,
    )


def test_json_to_namedtuple_deserializer_user_record(deserializer: Callable) -> None:
    record = deserializer(
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "create",'
        b' "_source": "test", "id": 1, "email": "test@example.com",'
        b' "score": "1.23", "is_active": true}'
    )

    assert record == UserRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.CREATE,
        ev_source="test",
        id=1,
        email="test@example.com",
        score=Decimal("1.23"),
        is_active=True,
    )


def test_json_to_namedtuple_deserializer_user_delete_record(
    deserializer: Callable,
) -> None:
    record = deserializer(
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "delete",'
        b' "_source": "test", "id": 1}'
    )

    assert record == UserDeleteRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.DELETE,
        ev_source="test",
        id=1,
    )


def test_json_to_namedtuple_deserializer_user_enumerate_record(
    deserializer: Callable,
) -> None:
    record = deserializer(
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "enumerate",'
        b' "_source": "test", "ids": [1, 2, 3]}'
    )

    assert record == UserEnumerateRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.ENUMERATE,
        ev_source="test",
        ids=[1, 2, 3],
    )


def test_json_to_namedtuple_deserializer_user_eos_record(
    deserializer: Callable,
) -> None:
    record = deserializer(
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "eos", "_source": "test"}'
    )

    assert record == UserEOSRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.EOS,
        ev_source="test",
    )



from .consumer import ConsumerWorker, ConsumerWorkerSettings
from .processor import (
    DataProcessor,
    DataProcessorSettings,
    DataTransformer,
    DataTransformerSettings,
)
from .source import DataSource, DataSourceSettings
from .types import DataType
from .worker import Worker, WorkerSettings
from .workerpool import WorkerDef, WorkerPool, WorkerPoolSettings

__version__ = "0.0.4"

__all__ = [
    "ConsumerWorker",
    "ConsumerWorkerSettings",
    "DataProcessor",
    "DataProcessorSettings",
    "DataSource",
    "DataSourceSettings",
    "DataType",
    "DataTransformer",
    "DataTransformerSettings",
    "Worker",
    "WorkerDef",
    "WorkerPool",
    "WorkerPoolSettings",
    "WorkerSettings",
]



import logging
from collections import defaultdict
from collections.abc import AsyncGenerator, Callable, Mapping
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property
from typing import Any, Generic, TypeAlias

from elasticsearch import AsyncElasticsearch
from elasticsearch.helpers import BulkIndexError, async_bulk

from ..types import DataType
from .base import DataWriter, DataWriterSettings
from .types import EnumerateIDsRecord, EventType, Version

log = logging.getLogger(__name__)

Document: TypeAlias = Mapping[str, Any]
Action: TypeAlias = dict[str, Any]


@dataclass
class DeleteByQuery:
    index: str
    query: Mapping[str, Any]


@dataclass
class ElasticsearchWriterSettings(Generic[DataType], DataWriterSettings[DataType]):
    elasticsearch: Callable[[], AsyncElasticsearch]
    index: str
    version_field: str
    record_serializer: Callable[[DataType], Document]
    process_eos: bool = False
    wait_for_completion: bool = False


class ElasticsearchWriter(Generic[DataType], DataWriter[DataType]):
    settings: ElasticsearchWriterSettings

    @cached_property
    def _es(self) -> AsyncElasticsearch:
        return self.settings.elasticsearch()

    def _es_index(self, record: DataType) -> str:
        return self.settings.index

    def _es_pipeline(self) -> str | None:
        return None

    def _es_id(self, record: DataType) -> str:
        return str(self.settings.id_getter(record))

    def _es_document(self, record: DataType) -> Document:
        return self.settings.record_serializer(record)

    def _es_version(self, record: DataType) -> int:
        version = self.settings.version_getter(record)
        if isinstance(version, datetime):
            version = int(version.timestamp())
        return version

    def _es_routing(self, record: DataType) -> str | None:
        return None

    def _action(self, event_type: EventType, record: DataType) -> Action:
        _index = self._es_index(record)
        _id = self._es_id(record)
        _version = self._es_version(record)
        _routing = self._es_routing(record)

        if event_type == EventType.DELETE:
            _op_type = "delete"
            _source = None
        else:
            _op_type = "index"
            _source = self._es_document(record)

        action = {
            "_op_type": _op_type,
            "_index": _index,
            "_id": _id,
            "_source": _source,
            "_version": _version,
            "_version_type": "external",
            "_routing": _routing,
        }
        return action

    def _enum_query(
        self, index: str, rec_ver: Version, enum_rec: EnumerateIDsRecord
    ) -> DeleteByQuery:
        assert (
            enum_rec.chunk is None
        ), "Enumerate messages with chunks is not implemented"

        query = {
            "bool": {
                "filter": [
                    {
                        "bool": {
                            "must_not": {
                                "ids": enum_rec.ids,
                            },
                        },
                    },
                    {
                        "range": {
                            self.settings.version_field: {
                                "lt": rec_ver,
                            },
                        },
                    },
                ],
            },
        }
        return DeleteByQuery(index=index, query=query)

    def _eos_query(self, index: str, rec_ver: Version) -> DeleteByQuery:
        query = {
            "bool": {
                "filter": [
                    {
                        "range": {
                            self.settings.version_field: {
                                "lt": rec_ver,
                            },
                        },
                    },
                ],
            },
        }
        return DeleteByQuery(index=index, query=query)

    async def _bulk_actions(self, actions: AsyncGenerator[Action]) -> None:
        pipeline = self._es_pipeline()

        try:
            await async_bulk(self._es, actions, pipeline=pipeline, raise_on_error=True)
        except BulkIndexError as error:
            errors_summary: defaultdict[str, int] = defaultdict(int)
            errors_samples: dict[str, Any] = {}
            for err in error.errors:
                if err.get("index"):
                    error_type = "index.%s" % (
                        err["index"].get("error", {}).get("type") or "unknown"
                    )
                elif err.get("delete"):
                    error_type = "delete.%s" % (
                        err["delete"].get("error", {}).get("type")
                        or err["delete"].get("result")
                        or "unknown"
                    )
                else:
                    error_type = "unknown"
                errors_summary[error_type] += 1
                errors_samples.setdefault(error_type, err)
            ignore_error_types = {
                "delete.not_found",
                "delete.version_conflict_engine_exception",
                "index.version_conflict_engine_exception",
            }
            if set(errors_summary.keys()) - ignore_error_types:
                log.error(
                    "Error(s) summery while inserting data to ElasticSearch:"
                    f" {dict(errors_summary)}",
                )
                for error_type, err in errors_samples.items():
                    log.error(
                        "Error while inserting data to ElasticSearch:"
                        f" {error_type}: {err}",
                    )

    async def _delete_by_query(self, query: DeleteByQuery) -> None:
        await self._es.delete_by_query(
            index=query.index,
            query=query.query,
            conflicts="proceed",
            refresh=True,
            wait_for_completion=self.settings.wait_for_completion,
        )
        if self.settings.wait_for_completion:
            await self._es.indices.refresh(index=query.index)

    async def process(self, batch: list[DataType]) -> None:
        indices: set[str] = set()

        async def _generator() -> AsyncGenerator[Action]:
            for record in batch:
                event_type = self.settings.event_type_getter(record)
                rec_ver = self.settings.version_getter(record)
                query: DeleteByQuery | None = None

                if event_type == EventType.ENUMERATE:
                    enum_rec = self.settings.enum_getter(record)
                    if enum_rec.chunk:
                        log.warning("Enumerate messages with chunks is not implemented")
                    else:
                        log.debug(
                            f"Accept enumerate message of version {rec_ver}"
                            f" with {len(enum_rec.ids)} object IDs"
                        )
                        index = self._es_index(record)
                        query = self._enum_query(index, rec_ver, enum_rec)
                elif event_type == EventType.EOS:
                    if self.settings.process_eos:
                        log.debug(f"Accept end of stream message of version {rec_ver}")
                        index = self._es_index(record)
                        query = self._eos_query(index, rec_ver)
                else:
                    action = self._action(event_type, record)
                    indices.add(action["_index"])
                    yield action

                if query:
                    await self._delete_by_query(query)

        actions = _generator()
        await self._bulk_actions(actions)
        if self.settings.wait_for_completion:
            for index in indices:
                await self._es.indices.refresh(index=index)



from collections.abc import Callable
from datetime import datetime, timezone
from decimal import Decimal

import pytest

from aiosafeconsumer.datasync import EventType

from .serializers import namedtuple_to_json_serializer
from .types import UserDeleteRecord, UserEnumerateRecord, UserEOSRecord, UserRecord


@pytest.fixture()
def serializer() -> Callable:
    return namedtuple_to_json_serializer()


def test_namedtuple_to_json_serializer_user_record(serializer: Callable) -> None:
    record = UserRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.CREATE,
        ev_source="test",
        id=1,
        email="test@example.com",
        score=Decimal("1.23"),
        is_active=True,
    )

    value = serializer(record)

    assert value == (
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "create",'
        b' "_source": "test", "id": 1, "email": "test@example.com",'
        b' "score": "1.23", "is_active": true}'
    )


def test_namedtuple_to_json_serializer_user_delete_record(serializer: Callable) -> None:
    record = UserDeleteRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.DELETE,
        ev_source="test",
        id=1,
    )

    value = serializer(record)

    assert value == (
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "delete",'
        b' "_source": "test", "id": 1}'
    )


def test_namedtuple_to_json_serializer_user_enumerate_record(
    serializer: Callable,
) -> None:
    record = UserEnumerateRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.ENUMERATE,
        ev_source="test",
        ids=[1, 2, 3],
    )

    value = serializer(record)

    assert value == (
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "enumerate",'
        b' "_source": "test", "ids": [1, 2, 3],'
        b' "chunk_session": null, "chunk_index": null, "total_chunks": null}'
    )


def test_namedtuple_to_json_serializer_user_eos_record(
    serializer: Callable,
) -> None:
    record = UserEOSRecord(
        ev_time=datetime(2024, 1, 1, tzinfo=timezone.utc),
        ev_type=EventType.EOS,
        ev_source="test",
    )

    value = serializer(record)

    assert value == (
        b'{"_time": "2024-01-01T00:00:00+00:00", "_type": "eos", "_source": "test"}'
    )



import asyncio
from datetime import datetime, timedelta, timezone
from decimal import Decimal

import pytest
from aiokafka import AIOKafkaProducer  # type: ignore
from elasticsearch import AsyncElasticsearch

from aiosafeconsumer import WorkerPool, WorkerPoolSettings
from aiosafeconsumer.datasync import EventType

from ..types import UserEOSRecord, UserRecord


@pytest.fixture
def ev_time() -> datetime:
    return datetime.now(timezone.utc).replace(microsecond=0)


@pytest.fixture
def users(ev_time: datetime) -> list[UserRecord]:
    ev_type = EventType.REFRESH
    ev_source = "test"
    return [
        UserRecord(
            ev_time=ev_time,
            ev_type=ev_type,
            ev_source=ev_source,
            id=1,
            email="user1@example.com",
            score=Decimal("1.23"),
            is_active=True,
        ),
        UserRecord(
            ev_time=ev_time,
            ev_type=ev_type,
            ev_source=ev_source,
            id=2,
            email="user2@example.com",
            score=Decimal("4.56"),
            is_active=False,
        ),
    ]


@pytest.mark.asyncio
async def test_on_empty_db(
    worker_pool_settings: WorkerPoolSettings,
    producer: AIOKafkaProducer,
    users: list[UserRecord],
    elasticsearch: AsyncElasticsearch,
    ev_time: datetime,
) -> None:
    pool = WorkerPool(worker_pool_settings, burst=True)
    task = asyncio.create_task(pool.run())

    await asyncio.sleep(0.1)
    await producer.start()
    try:
        for user in users:
            await producer.send("users", user)
    finally:
        await producer.flush()
        await producer.stop()

    await task

    results = await elasticsearch.search(
        index="users",
        query={
            "match_all": {},
        },
    )

    assert results["hits"]["hits"] == [
        {
            "_id": "1",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user1@example.com",
                "ev_source": "test",
                "ev_time": ev_time.isoformat(),
                "ev_type": "refresh",
                "id": 1,
                "is_active": True,
                "score": 1.23,
            },
        },
        {
            "_id": "2",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user2@example.com",
                "ev_source": "test",
                "ev_time": ev_time.isoformat(),
                "ev_type": "refresh",
                "id": 2,
                "is_active": False,
                "score": 4.56,
            },
        },
    ]


@pytest.mark.asyncio
async def test_update(
    worker_pool_settings: WorkerPoolSettings,
    producer: AIOKafkaProducer,
    users: list[UserRecord],
    elasticsearch: AsyncElasticsearch,
    ev_time: datetime,
) -> None:
    initial_users: list[UserRecord] = []
    for record in users:
        if record.id == 1:
            record = record._replace(ev_time=ev_time + timedelta(minutes=1))
        elif record.id == 2:
            record = record._replace(ev_time=ev_time - timedelta(minutes=1))
        initial_users.append(record)

    for record in initial_users:
        await elasticsearch.index(
            index="users",
            id=str(record.id),
            version=int(record.ev_time.timestamp()),
            version_type="external",
            document=record._asdict(),
        )

    pool = WorkerPool(worker_pool_settings, burst=True)
    task = asyncio.create_task(pool.run())

    await asyncio.sleep(0.1)
    await producer.start()
    try:
        for user in users:
            await producer.send("users", user)
    finally:
        await producer.flush()
        await producer.stop()

    await task

    results = await elasticsearch.search(
        index="users",
        query={
            "match_all": {},
        },
    )

    assert results["hits"]["hits"] == [
        {
            "_id": "1",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user1@example.com",
                "ev_source": "test",
                "ev_time": (ev_time + timedelta(minutes=1)).isoformat(),
                "ev_type": "refresh",
                "id": 1,
                "is_active": True,
                "score": 1.23,
            },
        },
        {
            "_id": "2",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user2@example.com",
                "ev_source": "test",
                "ev_time": ev_time.isoformat(),
                "ev_type": "refresh",
                "id": 2,
                "is_active": False,
                "score": 4.56,
            },
        },
    ]


@pytest.mark.asyncio
async def test_eos(
    worker_pool_settings: WorkerPoolSettings,
    producer: AIOKafkaProducer,
    users: list[UserRecord],
    elasticsearch: AsyncElasticsearch,
    ev_time: datetime,
) -> None:
    initial_users: list[UserRecord] = []
    for record in users:
        if record.id == 1:
            record = record._replace(
                id=11,
                ev_time=ev_time - timedelta(minutes=1),
            )
        elif record.id == 2:
            record = record._replace(
                id=12,
                ev_time=ev_time - timedelta(minutes=1),
            )
        initial_users.append(record)

    for record in initial_users:
        await elasticsearch.index(
            index="users",
            id=str(record.id),
            version=int(record.ev_time.timestamp()),
            version_type="external",
            document=record._asdict(),
        )

    pool = WorkerPool(worker_pool_settings, burst=True)
    task = asyncio.create_task(pool.run())

    await asyncio.sleep(0.1)
    await producer.start()
    try:
        for user in users:
            await producer.send("users", user)
        eos_record = UserEOSRecord(
            ev_time=ev_time,
            ev_type=EventType.EOS,
            ev_source="test",
        )
        await producer.send("users", eos_record)
    finally:
        await producer.flush()
        await producer.stop()

    await task

    await elasticsearch.indices.refresh(index="users")

    results = await elasticsearch.search(
        index="users",
        query={
            "match_all": {},
        },
    )

    assert results["hits"]["hits"] == [
        {
            "_id": "1",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user1@example.com",
                "ev_source": "test",
                "ev_time": ev_time.isoformat(),
                "ev_type": "refresh",
                "id": 1,
                "is_active": True,
                "score": 1.23,
            },
        },
        {
            "_id": "2",
            "_index": "users",
            "_score": 1.0,
            "_source": {
                "email": "user2@example.com",
                "ev_source": "test",
                "ev_time": ev_time.isoformat(),
                "ev_type": "refresh",
                "id": 2,
                "is_active": False,
                "score": 4.56,
            },
        },
    ]



import os

import pytest
import pytest_asyncio
from aiokafka import AIOKafkaProducer  # type: ignore

from .serializers import namedtuple_to_json_serializer


@pytest.fixture
def kafka_bootstrap_servers() -> list[str]:
    return os.getenv("KAFKA_BOOTSTRAP_SERVERS", "").split(",")


@pytest_asyncio.fixture
async def producer(kafka_bootstrap_servers: list[str]) -> AIOKafkaProducer:
    producer = AIOKafkaProducer(
        bootstrap_servers=kafka_bootstrap_servers,
        value_serializer=namedtuple_to_json_serializer(),
    )
    return producer



import json
import types
import typing
from collections.abc import Callable
from datetime import datetime
from decimal import Decimal
from typing import NamedTuple

from .constants import ENCODING, FIELD_TRANSLATION

_FIELDS_TR = dict([(obj_f, bus_f) for bus_f, obj_f in FIELD_TRANSLATION])


def namedtuple_to_json_serializer(
    encoding: str = ENCODING,
    ensure_ascii: bool = False,
) -> Callable[[NamedTuple], bytes]:
    def serializer(obj: NamedTuple) -> bytes:
        payload = obj._asdict()

        for f in obj._fields:
            f_type = obj.__annotations__[f]
            if typing.get_origin(f_type) is types.UnionType:
                f_type = typing.get_args(f_type)[0]

            v = payload[f]
            if f_type is datetime:
                if isinstance(v, datetime):
                    payload[f] = v.isoformat()
            if f_type is Decimal:
                if isinstance(v, Decimal):
                    payload[f] = str(v)

        transformed = {_FIELDS_TR.get(k, k): v for k, v in payload.items()}
        return json.dumps(transformed, ensure_ascii=ensure_ascii).encode(encoding)

    return serializer



from collections.abc import Generator
from datetime import timedelta
from unittest import mock

import pytest
from aiokafka import ConsumerRecord, TopicPartition  # type: ignore
from aiokafka.errors import KafkaConnectionError, KafkaError  # type: ignore

from aiosafeconsumer.kafka import (
    InformativeRebalanceListener,
    KafkaSource,
    KafkaSourceSettings,
)


class StrKafkaSourceSettings(KafkaSourceSettings[str]):
    pass


class StrKafkaSource(KafkaSource[str]):
    settings: StrKafkaSourceSettings


@pytest.fixture
def kafka_source_settings() -> KafkaSourceSettings:
    def value_deserializer(value: bytes) -> str:
        return value.decode()

    settings = StrKafkaSourceSettings(
        topics=["test-topic"],
        bootstrap_servers=["localhost:9092"],
        value_deserializer=value_deserializer,
        max_consumer_errors=1,
        sleep_on_consumer_error=timedelta(seconds=0.1),
    )
    return settings


@pytest.fixture
def kafka_source(kafka_source_settings: KafkaSourceSettings) -> KafkaSource:
    source = StrKafkaSource(kafka_source_settings)
    return source


@pytest.fixture
def data() -> list[str]:
    return [str(i) for i in range(20)]


@pytest.fixture
def kafka_consumer_mock() -> Generator[mock.Mock]:
    with mock.patch("aiosafeconsumer.kafka.AIOKafkaConsumer") as consumer_class_m:
        consumer_m = consumer_class_m.return_value
        consumer_m.start = mock.AsyncMock()
        consumer_m.stop = mock.AsyncMock()
        consumer_m.getmany = mock.AsyncMock()
        consumer_m.commit = mock.AsyncMock()
        consumer_m._coordinator = mock.AsyncMock()

        yield consumer_m


@pytest.fixture
def kafka_consumer_data(kafka_consumer_mock: mock.Mock, data: list[str]) -> None:
    kafka_consumer_mock.getmany.return_value = {
        TopicPartition("test-topic", 0): [
            ConsumerRecord(
                **{
                    **{f: None for f in ConsumerRecord.__dataclass_fields__.keys()},
                    "topic": "test-topic",
                    "partition": 0,
                    "offset": 0,
                    "value": value,
                },
            )
            for value in data
        ],
    }


@pytest.mark.asyncio
async def test_kafka_source_read(
    kafka_consumer_data: None, data: list[str], kafka_source: StrKafkaSource
) -> None:
    generator = kafka_source.read()
    batch = await anext(generator)
    await generator.aclose()

    assert batch == data


@pytest.mark.asyncio
async def test_kafka_source_read_with_start_errors(
    kafka_consumer_mock: mock.Mock,
    kafka_consumer_data: None,
    data: list[str],
    kafka_source: StrKafkaSource,
) -> None:
    async def start_se() -> None:
        if kafka_consumer_mock.start.call_count < 3:
            raise KafkaConnectionError("Test error")

    kafka_consumer_mock.start.side_effect = start_se
    generator = kafka_source.read()
    batch = await anext(generator)
    await generator.aclose()

    assert batch == data
    assert kafka_consumer_mock.stop.call_count == 1
    assert kafka_consumer_mock.start.call_count == 3


@pytest.mark.asyncio
async def test_kafka_source_read_with_commit_errors(
    kafka_consumer_mock: mock.Mock,
    kafka_consumer_data: None,
    data: list[str],
    kafka_source: StrKafkaSource,
) -> None:
    async def commit_se(offsets: dict[TopicPartition, int]) -> None:
        if kafka_consumer_mock.commit.call_count < 3:
            raise KafkaError("Test error")

    kafka_consumer_mock.commit.side_effect = commit_se
    generator = kafka_source.read()
    batch = await anext(generator)
    await kafka_source.commit(batch)
    await anext(generator)
    await kafka_source.commit(batch)
    await anext(generator)
    await kafka_source.commit(batch)
    await anext(generator)
    await generator.aclose()

    assert batch == data
    assert kafka_consumer_mock.commit.call_count == 3


@pytest.mark.asyncio
async def test_kafka_source_str(kafka_source: StrKafkaSource) -> None:
    assert str(kafka_source) == "StrKafkaSource"


@pytest.mark.asyncio
async def test_informative_rebalance_listener() -> None:
    listener = InformativeRebalanceListener()
    assert listener.revoked == set()
    assert listener.assigned == set()

    await listener.on_partitions_revoked([TopicPartition("test-topic", 0)])
    await listener.on_partitions_assigned([TopicPartition("test-topic", 1)])

    assert listener.revoked == {TopicPartition("test-topic", 0)}
    assert listener.assigned == {TopicPartition("test-topic", 1)}



import logging
import pickle
from collections.abc import AsyncGenerator, Callable, Sequence
from dataclasses import dataclass
from datetime import timedelta
from typing import Generic, TypeAlias, cast, no_type_check

from redis.asyncio import Redis

from ..types import DataType
from .base import DataWriter, DataWriterSettings
from .types import EnumerateIDsChunk, EnumerateIDsRecord, EventType, ObjectID, Version

log = logging.getLogger(__name__)


StrID: TypeAlias = str


@dataclass
class RedisWriterSettings(Generic[DataType], DataWriterSettings[DataType]):
    redis: Callable[[], Redis]
    version_serializer: Callable[[Version], bytes]
    version_deserializer: Callable[[bytes], Version]
    record_serializer: Callable[[DataType], bytes]
    key_prefix: str
    versions_key: str
    enum_chunks_key_prefix: str | None = None
    process_eos: bool = False
    versions_expire: timedelta = timedelta(hours=24)
    record_expire: timedelta = timedelta(hours=30)


class RedisWriter(Generic[DataType], DataWriter[DataType]):
    settings: RedisWriterSettings

    def _obj_key(self, obj_id: StrID) -> bytes:
        return f"{self.settings.key_prefix}{obj_id}".encode()

    def _obj_version_key(self, obj_id: StrID) -> bytes:
        return obj_id.encode()

    def _versions_key(self) -> bytes:
        return self.settings.versions_key.encode()

    def _enum_chunk_key_prefix(self) -> str:
        return (
            self.settings.enum_chunks_key_prefix
            or f"{self.settings.versions_key}_chunk:"
        )

    def _enum_chunk_key(self, chunk: EnumerateIDsChunk) -> bytes:
        prefix = self._enum_chunk_key_prefix()
        key = f"{prefix}{chunk.session}:{chunk.chunk_index}"
        return key.encode()

    def _enum_chunk_match_pattern(self, session: str) -> bytes:
        prefix = self._enum_chunk_key_prefix()
        key = f"{prefix}{session}:*"
        return key.encode()

    @no_type_check
    def _compare_versions(self, a: Version, b: Version) -> int:
        assert type(a) is type(b)
        if a < b:
            return -1
        if a > b:
            return 1
        return 0

    async def _get_versions(self, redis: Redis) -> dict[StrID, Version]:
        raw_versions = await redis.hgetall(self._versions_key())

        versions: dict[StrID, Version] = {
            raw_id.decode(): self.settings.version_deserializer(raw_ver)
            for raw_id, raw_ver in raw_versions.items()
        }
        return versions

    async def _save_enum_chunk(
        self, redis: Redis, chunk: EnumerateIDsChunk, ids: Sequence[ObjectID]
    ) -> None:
        key = self._enum_chunk_key(chunk)
        ids_set = set(str(id_) for id_ in ids)
        value = pickle.dumps(ids_set)
        await redis.set(key, value, ex=self.settings.versions_expire)

    async def _iter_enum_chunks(
        self, redis: Redis, session: str
    ) -> AsyncGenerator[tuple[int, set[StrID]]]:
        match = self._enum_chunk_match_pattern(session)
        async for key in redis.scan_iter(match):
            chunk_index = int(key.decode().rsplit(":")[-1])
            value = await redis.get(key)
            if value:
                ids = cast(set[StrID], pickle.loads(value))
                yield (chunk_index, ids)

    async def process(self, batch: list[DataType]) -> None:
        settings = self.settings
        redis = settings.redis()

        enum_records: dict[Version, list[EnumerateIDsRecord]] = {}
        eos_versions: set[Version] = set()
        upsert_records: dict[StrID, DataType] = {}
        upsert_versions: dict[StrID, Version] = {}

        versions = await self._get_versions(redis)
        log.debug(f"Loaded {len(versions)} object versions from Redis")

        for record in batch:
            event_type = settings.event_type_getter(record)
            rec_ver = settings.version_getter(record)

            if event_type == EventType.ENUMERATE:
                enum_rec = settings.enum_getter(record)
                enum_records.setdefault(rec_ver, [])
                enum_records[rec_ver].append(enum_rec)
            elif event_type == EventType.EOS:
                if self.settings.process_eos:
                    eos_versions.add(rec_ver)
            else:
                obj_id = str(settings.id_getter(record))
                cur_ver = upsert_versions.get(obj_id, versions.get(obj_id))
                if cur_ver is None or self._compare_versions(rec_ver, cur_ver) > 0:
                    upsert_records[obj_id] = record
                    upsert_versions[obj_id] = rec_ver

        update_records: dict[bytes, bytes] = {}
        update_versions: dict[bytes, bytes] = {}
        del_versions: list[bytes] = []
        del_records: list[bytes] = []

        for obj_id, record in upsert_records.items():
            event_type = settings.event_type_getter(record)
            ver_key = self._obj_version_key(obj_id)
            key = self._obj_key(obj_id)
            if event_type == EventType.DELETE:
                del_versions.append(ver_key)
                del_records.append(key)
                continue
            value = settings.record_serializer(record)
            update_records[key] = value
            ver_value = settings.version_serializer(upsert_versions[obj_id])
            update_versions[ver_key] = ver_value

        enum_to_process: list[tuple[Version, set[StrID]]] = []

        for rec_ver, rec_list in enum_records.items():
            for rec in rec_list:
                if rec.chunk is None:
                    enum_ids: set[StrID] = set(str(id_) for id_ in rec.ids)
                    enum_to_process.append((rec_ver, enum_ids))
                    continue

                chunks_in_db: set[int] = set()
                ids_in_db: set[StrID] = set()

                async for chunk_index, ids in self._iter_enum_chunks(
                    redis, rec.chunk.session
                ):
                    chunks_in_db.add(chunk_index)
                    ids_in_db |= ids

                complete_chunks = set(range(rec.chunk.total_chunks))
                if chunks_in_db | {rec.chunk.chunk_index} == complete_chunks:
                    enum_ids = ids_in_db | set(str(id_) for id_ in rec.ids)
                    enum_to_process.append((rec_ver, enum_ids))
                    for chunk_index in complete_chunks:
                        del_records.append(
                            self._enum_chunk_key(
                                EnumerateIDsChunk(
                                    session=rec.chunk.session,
                                    chunk_index=chunk_index,
                                    total_chunks=rec.chunk.total_chunks,
                                )
                            )
                        )
                else:
                    await self._save_enum_chunk(redis, rec.chunk, rec.ids)

        for enum_ver, enum_ids in enum_to_process:
            outdated_ids = set(
                obj_id
                for obj_id, rec_ver in versions.items()
                if self._compare_versions(rec_ver, enum_ver) < 0
            )
            fresh_ids = set(
                obj_id
                for obj_id, rec_ver in upsert_versions.items()
                if self._compare_versions(rec_ver, enum_ver) >= 0
            )
            ids_to_delete = outdated_ids - enum_ids - fresh_ids
            log.debug(
                f"Accept enumerate message of version {enum_ver}"
                f" with {len(enum_ids)} object IDs,"
                f" {len(ids_to_delete)} records will be deleted"
            )
            del_records.extend([self._obj_key(obj_id) for obj_id in ids_to_delete])
            del_versions.extend(
                [self._obj_version_key(obj_id) for obj_id in ids_to_delete]
            )

        for eos_ver in eos_versions:
            outdated_ids = set(
                obj_id
                for obj_id, rec_ver in versions.items()
                if self._compare_versions(rec_ver, eos_ver) < 0
            )
            fresh_ids = set(
                obj_id
                for obj_id, rec_ver in upsert_versions.items()
                if self._compare_versions(rec_ver, eos_ver) >= 0
            )
            ids_to_delete = outdated_ids - fresh_ids
            log.debug(
                f"Accept end of stream message of version {eos_ver}, "
                f" {len(ids_to_delete)} records will be deleted"
            )
            del_records.extend([self._obj_key(obj_id) for obj_id in ids_to_delete])
            del_versions.extend(
                [self._obj_version_key(obj_id) for obj_id in ids_to_delete]
            )

        async with redis.pipeline(transaction=True) as pipe:
            if update_records:
                pipe.mset(update_records)  # type: ignore
            for key in update_records.keys():
                pipe.expire(key, settings.record_expire)
            if del_records:
                pipe.delete(*del_records)
            if del_versions:
                pipe.hdel(self._versions_key(), *del_versions)
            if update_versions:
                pipe.hset(self._versions_key(), mapping=update_versions)  # type: ignore
            pipe.expire(self._versions_key(), settings.versions_expire)
            await pipe.execute()

        log.debug(
            f"{len(update_records)} records was added/updated and {len(del_records)}"
            " records was deleted"
        )



import os
from collections.abc import AsyncGenerator
from datetime import timedelta
from typing import Any

import pytest
import pytest_asyncio
from elasticsearch import AsyncElasticsearch

from aiosafeconsumer import WorkerDef, WorkerPoolSettings

from ..deserializers import json_to_namedtuple_deserializer
from ..sources import UsersKafkaSource, UsersKafkaSourceSettings
from ..types import UserDeleteRecord, UserEnumerateRecord, UserEOSRecord, UserRecord
from ..workers import UsersWorker, UsersWorkerSettings
from .processors import UsersElasticsearchWriter, UsersElasticsearchWriterSettings


@pytest.fixture
def elasticsearch_url() -> str:
    return os.getenv("ELASTICSEARCH_URL", "")


@pytest.fixture
def elasticsearch_indices() -> dict[str, dict[str, Any]]:
    return {
        "users": {
            "mappings": {
                "properties": {
                    "ev_time": {"type": "date"},
                    "ev_type": {"type": "keyword"},
                    "ev_source": {"type": "keyword"},
                    "id": {"type": "integer"},
                    "email": {"type": "keyword"},
                    "score": {"type": "float"},
                    "is_active": {"type": "boolean"},
                },
            },
        },
    }


@pytest_asyncio.fixture
async def elasticsearch(
    elasticsearch_url: str,
    elasticsearch_indices: dict[str, dict[str, Any]],
) -> AsyncGenerator[AsyncElasticsearch]:
    es = AsyncElasticsearch(elasticsearch_url)
    try:
        for index_name, index_def in elasticsearch_indices.items():
            await es.indices.create(index=index_name, **index_def)
        yield es
    finally:
        for index_name in elasticsearch_indices.keys():
            await es.indices.delete(index=index_name, ignore_unavailable=True)
        await es.close()


@pytest.fixture
def worker_pool_settings(
    elasticsearch: AsyncElasticsearch,
    kafka_bootstrap_servers: list[str],
) -> WorkerPoolSettings:
    pool_settings = WorkerPoolSettings(
        workers=[
            WorkerDef(
                worker_class=UsersWorker,
                worker_settings=UsersWorkerSettings(
                    source_class=UsersKafkaSource,
                    source_settings=UsersKafkaSourceSettings(
                        topics=["users"],
                        bootstrap_servers=kafka_bootstrap_servers,
                        group_id="sync_users",
                        value_deserializer=json_to_namedtuple_deserializer(
                            UserRecord,
                            UserDeleteRecord,
                            UserEnumerateRecord,
                            UserEOSRecord,
                        ),
                        getmany_timeout=timedelta(seconds=0.1),
                        kwargs={
                            "auto_offset_reset": "earliest",
                            "fetch_max_wait_ms": 100,
                        },
                    ),
                    processor_class=UsersElasticsearchWriter,
                    processor_settings=UsersElasticsearchWriterSettings(
                        elasticsearch=lambda: elasticsearch,
                        index="users",
                        version_field="ev_time",
                        process_eos=True,
                        wait_for_completion=True,
                    ),
                ),
            ),
        ],
    )
    return pool_settings



import json
import types
import typing
from collections.abc import Callable
from datetime import datetime
from decimal import Decimal
from typing import NamedTuple, TypeVar

from aiosafeconsumer.datasync import EventType

from .constants import ENCODING, FIELD_TRANSLATION

_FIELDS_TR = dict(FIELD_TRANSLATION)

RecordT = TypeVar("RecordT", bound=NamedTuple)
DeleteRecordT = TypeVar("DeleteRecordT", bound=NamedTuple)
EnumerateRecordT = TypeVar("EnumerateRecordT", bound=NamedTuple)
EOSRecordT = TypeVar("EOSRecordT", bound=NamedTuple)


def json_to_namedtuple_deserializer(
    record_class: type[RecordT],
    delete_record_class: type[DeleteRecordT] | None = None,
    enumerate_record_class: type[EnumerateRecordT] | None = None,
    eos_record_class: type[EOSRecordT] | None = None,
    encoding: str = ENCODING,
) -> Callable[[bytes], RecordT | DeleteRecordT | EnumerateRecordT | EOSRecordT]:
    def deserializer(
        value: bytes,
    ) -> RecordT | DeleteRecordT | EnumerateRecordT | EOSRecordT:
        payload = json.loads(value.decode(encoding))
        ev_type = EventType(payload["_type"])

        class_: type[NamedTuple] = record_class
        if ev_type == EventType.DELETE:
            assert delete_record_class is not None
            class_ = delete_record_class
        elif ev_type == EventType.ENUMERATE:
            assert enumerate_record_class is not None
            class_ = enumerate_record_class
        elif ev_type == EventType.EOS:
            assert eos_record_class is not None
            class_ = eos_record_class

        fields = set(f for f in class_._fields)
        transformed = [(_FIELDS_TR.get(k, k), v) for k, v in payload.items()]
        data = {k: v for k, v in transformed if k in fields}

        for f in class_._fields:
            f_type = class_.__annotations__[f]
            if typing.get_origin(f_type) is types.UnionType:
                f_type_args = typing.get_args(f_type)
                f_type = f_type_args[0]

            if f in data:
                v = data[f]
            elif f in class_._field_defaults:
                v = class_._field_defaults[f]
            else:
                raise ValueError(f"No required field: {f}")

            if f_type is datetime:
                if type(v) is int:
                    data[f] = datetime.fromtimestamp(v)
                elif type(v) is str:
                    data[f] = datetime.fromisoformat(v)
            if f_type is Decimal:
                if type(v) is str or type(v) is int:
                    data[f] = Decimal(v)
            elif f_type is EventType:
                if type(v) is str:
                    data[f] = EventType(v)

        return class_(**data)  # type: ignore

    return deserializer



from .base import DataWriter, DataWriterSettings
from .types import EnumerateIDsChunk, EnumerateIDsRecord, EventType, ObjectID, Version

__all__ = [
    "EnumerateIDsChunk",
    "EnumerateIDsRecord",
    "EventType",
    "DataWriter",
    "DataWriterSettings",
    "EventType",
    "ObjectID",
    "Version",
]



import json
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from datetime import timedelta
from fnmatch import fnmatch
from typing import NamedTuple, cast
from unittest import mock

import pytest
from redis.asyncio import Redis
from redis.asyncio.client import Pipeline

from aiosafeconsumer.datasync import (
    EnumerateIDsChunk,
    EnumerateIDsRecord,
    EventType,
    ObjectID,
    Version,
)
from aiosafeconsumer.datasync.redis import RedisWriter, RedisWriterSettings

KEY_PREFIX = "item:"
VERSIONS_KEY = "items"
ENUM_CHUNKS_PREFIX = "items_chunk:"


class ItemRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int
    data: str


class ItemDeleteRecord(NamedTuple):
    ev_type: EventType
    version: int
    id: int


class ItemEnumerateRecord(NamedTuple):
    ev_type: EventType
    version: int
    ids: list[int]
    chunk_session: str | None = None
    chunk_index: int | None = None
    total_chunks: int | None = None


class ItemEOSRecord(NamedTuple):
    ev_type: EventType
    version: int


Item = ItemRecord | ItemDeleteRecord | ItemEnumerateRecord | ItemEOSRecord


@dataclass
class Settings(RedisWriterSettings[Item]):
    pass


class Writer(RedisWriter[Item]):
    settings: Settings


@pytest.fixture
def redis_mock() -> Redis:
    versions: dict[bytes, bytes] = {}
    objects: dict[bytes, bytes] = {}
    expire: dict[bytes, timedelta] = {}
    enum_chunks: dict[bytes, bytes] = {}

    async def hgetall(key: bytes) -> dict[bytes, bytes]:
        if key == VERSIONS_KEY.encode():
            return versions.copy()
        return {}

    def hdel(name: bytes, *keys: bytes) -> None:
        if name == VERSIONS_KEY.encode():
            for k in keys:
                versions.pop(k, None)

    def hset(name: bytes, mapping: dict[bytes, bytes]) -> None:
        if name == VERSIONS_KEY.encode():
            versions.update(mapping)

    def mset(mapping: dict[bytes, bytes]) -> None:
        new_objects: dict[bytes, bytes] = {}
        new_enum_chunks: dict[bytes, bytes] = {}

        for key, value in mapping.items():
            if key.decode().startswith(ENUM_CHUNKS_PREFIX):
                new_enum_chunks[key] = value
            else:
                new_objects[key] = value
        objects.update(new_objects)
        enum_chunks.update(new_enum_chunks)

    def delete(*keys: bytes) -> None:
        for key in keys:
            key_str = key.decode()
            if key_str.startswith(ENUM_CHUNKS_PREFIX):
                enum_chunks.pop(key, None)
            else:
                objects.pop(key, None)
            expire.pop(key, None)

    def expire_(key: bytes, ex: timedelta) -> None:
        expire[key] = ex

    async def get(key: bytes) -> bytes | None:
        if key.decode().startswith(ENUM_CHUNKS_PREFIX):
            return enum_chunks.get(key)
        return objects.get(key)

    async def set_(key: bytes, value: bytes, ex: timedelta | None = None) -> None:
        if key.decode().startswith(ENUM_CHUNKS_PREFIX):
            enum_chunks[key] = value
        else:
            objects[key] = value
        if ex:
            expire[key] = ex

    async def scan_iter(match: bytes) -> AsyncGenerator[bytes]:
        match_str = match.decode()
        if match_str.startswith(ENUM_CHUNKS_PREFIX):
            for key in enum_chunks.keys():
                key_str = key.decode()
                if fnmatch(key_str, match_str):
                    yield key
        elif match_str.startswith(KEY_PREFIX):
            for key in objects.keys():
                key_str = key.decode()
                if fnmatch(key_str, match_str):
                    yield key

    redis = mock.MagicMock(spec=Redis)
    redis.hgetall.side_effect = hgetall
    redis.get.side_effect = get
    redis.set.side_effect = set_
    redis.scan_iter.side_effect = scan_iter

    pipe = mock.MagicMock(spec=Pipeline)
    redis.pipeline.return_value.__aenter__.return_value = pipe

    pipe.delete.side_effect = delete
    pipe.expire.side_effect = expire_
    pipe.hdel.side_effect = hdel
    pipe.hset.side_effect = hset
    pipe.mset.side_effect = mset

    redis._test_versions = versions
    redis._test_objects = objects
    redis._test_expire = expire

    return redis


@pytest.fixture
def settings(redis_mock: Redis) -> Settings:
    ENCODING = "utf-8"

    def get_redis() -> Redis:
        return redis_mock

    def version_getter(item: Item) -> Version:
        return item.version

    def record_serializer(item: Item) -> bytes:
        return json.dumps(item._asdict()).encode(ENCODING)

    def event_type_getter(item: Item) -> EventType:
        return item.ev_type

    def id_getter(item: Item) -> ObjectID:
        assert isinstance(item, ItemRecord) or isinstance(item, ItemDeleteRecord)
        return item.id

    def enum_getter(item: Item) -> EnumerateIDsRecord:
        assert isinstance(item, ItemEnumerateRecord)
        ids = cast(list[ObjectID], item.ids)
        chunk: EnumerateIDsChunk | None = None
        if item.chunk_session:
            assert item.chunk_index is not None
            assert item.total_chunks is not None
            chunk = EnumerateIDsChunk(
                chunk_index=item.chunk_index,
                total_chunks=item.total_chunks,
                session=item.chunk_session,
            )
        return EnumerateIDsRecord(ids=ids, chunk=chunk)

    def version_serializer(ver: Version) -> bytes:
        return str(ver).encode()

    def version_deserializer(val: bytes) -> Version:
        return int(val.decode())

    settings = Settings(
        redis=get_redis,
        version_getter=version_getter,
        version_serializer=version_serializer,
        version_deserializer=version_deserializer,
        record_serializer=record_serializer,
        key_prefix=KEY_PREFIX,
        versions_key=VERSIONS_KEY,
        event_type_getter=event_type_getter,
        id_getter=id_getter,
        enum_getter=enum_getter,
    )
    return settings


@pytest.mark.asyncio
async def test_redis_writer_initial_empty(
    redis_mock: Redis, settings: Settings
) -> None:
    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=1, id=3, data="3v1"),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"1",
        b"2": b"2",
        b"3": b"1",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:1": b'{"ev_type": "create", "version": 1, "id": 1, "data": "1v1"}',
        b"item:2": b'{"ev_type": "update", "version": 2, "id": 2, "data": "2v2"}',
        b"item:3": b'{"ev_type": "refresh", "version": 1, "id": 3, "data": "3v1"}',
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"item:1": timedelta(hours=30),
        b"item:2": timedelta(hours=30),
        b"item:3": timedelta(hours=30),
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_upsert(redis_mock: Redis, settings: Settings) -> None:
    redis_mock._test_versions.update(  # type: ignore
        {
            b"2": b"3",
            b"3": b"1",
            b"4": b"0",
            b"5": b"2",
        },
    )
    redis_mock._test_objects.update(  # type:ignore
        {
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
            b"item:5": b"",
        },
    )

    items: list[Item] = [
        ItemRecord(ev_type=EventType.CREATE, version=1, id=1, data="1v1"),
        ItemRecord(ev_type=EventType.UPDATE, version=2, id=2, data="2v2"),
        ItemRecord(ev_type=EventType.REFRESH, version=2, id=3, data="3v2"),
        ItemDeleteRecord(ev_type=EventType.DELETE, version=1, id=4),
        ItemDeleteRecord(ev_type=EventType.DELETE, version=1, id=5),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"1",
        b"2": b"3",
        b"3": b"2",
        b"5": b"2",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:1": b'{"ev_type": "create", "version": 1, "id": 1, "data": "1v1"}',
        b"item:2": b"",
        b"item:3": b'{"ev_type": "refresh", "version": 2, "id": 3, "data": "3v2"}',
        b"item:5": b"",
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"item:1": timedelta(hours=30),
        b"item:3": timedelta(hours=30),
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_enumerate(redis_mock: Redis, settings: Settings) -> None:
    redis_mock._test_versions.update(  # type: ignore
        {
            b"1": b"0",
            b"2": b"2",
            b"3": b"1",
            b"4": b"0",
        },
    )
    redis_mock._test_objects.update(  # type:ignore
        {
            b"item:1": b"",
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
        },
    )

    items: list[Item] = [
        ItemEnumerateRecord(ev_type=EventType.ENUMERATE, version=1, ids=[1, 3]),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"0",
        b"2": b"2",
        b"3": b"1",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:1": b"",
        b"item:2": b"",
        b"item:3": b"",
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_enumerate_with_chunks(
    redis_mock: Redis, settings: Settings
) -> None:
    redis_mock._test_versions.update(  # type: ignore
        {
            b"1": b"0",
            b"2": b"0",
            b"3": b"0",
            b"4": b"0",
            b"5": b"0",
            b"6": b"0",
        },
    )
    redis_mock._test_objects.update(  # type:ignore
        {
            b"item:1": b"",
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
            b"item:5": b"",
            b"item:6": b"",
        },
    )

    writer = Writer(settings)

    items: list[Item] = [
        ItemEnumerateRecord(
            ev_type=EventType.ENUMERATE,
            version=1,
            ids=[1, 3],
            chunk_session="session",
            chunk_index=0,
            total_chunks=2,
        ),
    ]
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"0",
        b"2": b"0",
        b"3": b"0",
        b"4": b"0",
        b"5": b"0",
        b"6": b"0",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:1": b"",
        b"item:2": b"",
        b"item:3": b"",
        b"item:4": b"",
        b"item:5": b"",
        b"item:6": b"",
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"items": timedelta(hours=24),
        b"items_chunk:session:0": timedelta(hours=24),
    }

    items = [
        ItemEnumerateRecord(
            ev_type=EventType.ENUMERATE,
            version=1,
            ids=[4, 6],
            chunk_session="session",
            chunk_index=1,
            total_chunks=2,
        ),
    ]
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"1": b"0",
        b"3": b"0",
        b"4": b"0",
        b"6": b"0",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:1": b"",
        b"item:3": b"",
        b"item:4": b"",
        b"item:6": b"",
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"items": timedelta(hours=24),
    }


@pytest.mark.asyncio
async def test_redis_writer_eos(redis_mock: Redis, settings: Settings) -> None:
    settings.process_eos = True

    redis_mock._test_versions.update(  # type: ignore
        {
            b"1": b"0",
            b"2": b"2",
            b"3": b"1",
            b"4": b"0",
        },
    )
    redis_mock._test_objects.update(  # type:ignore
        {
            b"item:1": b"",
            b"item:2": b"",
            b"item:3": b"",
            b"item:4": b"",
        },
    )

    items: list[Item] = [
        ItemEOSRecord(ev_type=EventType.EOS, version=1),
    ]

    writer = Writer(settings)
    await writer.process(items)

    assert redis_mock._test_versions == {  # type: ignore
        b"2": b"2",
        b"3": b"1",
    }
    assert redis_mock._test_objects == {  # type: ignore
        b"item:2": b"",
        b"item:3": b"",
    }
    assert redis_mock._test_expire == {  # type: ignore
        b"items": timedelta(hours=24),
    }



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Generic

from .types import DataType, TargetDataType

log = logging.getLogger(__name__)


@dataclass
class DataProcessorSettings:
    pass


class DataProcessor(Generic[DataType], ABC):
    def __init__(self, settings: DataProcessorSettings | None = None):
        self.settings = settings or DataProcessorSettings()

    def __str__(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    async def process(self, batch: list[DataType]) -> None:  # pragma: no cover
        pass


@dataclass
class DataTransformerSettings(Generic[DataType, TargetDataType], DataProcessorSettings):
    target_processor_class: type[DataProcessor[TargetDataType]]
    target_processor_settings: DataProcessorSettings


class DataTransformer(Generic[DataType, TargetDataType], DataProcessor[DataType], ABC):
    def __init__(self, settings: DataTransformerSettings) -> None:
        self.target_processor = settings.target_processor_class(
            settings.target_processor_settings
        )

    async def process(self, batch: list[DataType]) -> None:
        new_batch = await self.transform(batch)
        log.info(
            f"{len(batch)} items was transformed by {self} to {len(new_batch)} items"
        )
        await self.target_processor.process(new_batch)

    @abstractmethod
    async def transform(
        self, batch: list[DataType]
    ) -> list[TargetDataType]:  # pragma: no cover
        pass



from abc import ABC, abstractmethod
from collections.abc import AsyncGenerator
from dataclasses import dataclass
from typing import Generic

from .types import DataType


@dataclass
class DataSourceSettings:
    pass


class DataSource(Generic[DataType], ABC):
    """
    Abstract class to represent source of data.
    """

    def __init__(self, settings: DataSourceSettings | None = None):
        self.settings = settings or DataSourceSettings()

    def __str__(self) -> str:
        return self.__class__.__name__

    @abstractmethod
    async def read(self) -> AsyncGenerator[list[DataType], None]:  # pragma: no cover
        """
        Returns async generator to read batches of items.
        """
        if False:
            batch: list[DataType] = []
            yield batch
        raise NotImplementedError

    async def commit(self, batch: list[DataType]) -> None:
        """
        This method is called from worker just after batch of items was processed.

        """
        pass



import asyncio
from typing import cast

import pytest

from aiosafeconsumer import ConsumerWorker, DataTransformer

from .conftest import IntProcessor, StrProcessor, StrSource


@pytest.mark.asyncio
async def test_consumer(consumer: ConsumerWorker) -> None:
    source = cast(StrSource, consumer.source)
    processor = cast(StrProcessor, consumer.processor)

    assert processor.storage == []

    items_count = sum([len(batch) for batch in source.BATCHES])

    task = asyncio.create_task(consumer.run())

    while not task.done() and len(processor.storage) < items_count * 3:
        await asyncio.sleep(0)

    assert not task.done()
    task.cancel()

    assert processor.storage == [
        *source.BATCHES[0],
        *source.BATCHES[1],
        *source.BATCHES[0],
        *source.BATCHES[1],
        *source.BATCHES[0],
        *source.BATCHES[1],
    ]


@pytest.mark.asyncio
async def test_consumer_source_stop(consumer: ConsumerWorker) -> None:
    source = cast(StrSource, consumer.source)
    source.settings.stop_on = 2
    processor = cast(StrProcessor, consumer.processor)

    assert processor.storage == []

    items_count = sum([len(batch) for batch in source.BATCHES])

    task = asyncio.create_task(consumer.run())

    while not task.done() and len(processor.storage) < items_count * 10:
        await asyncio.sleep(0)

    assert task.done()

    assert processor.storage == [
        *source.BATCHES[0],
        *source.BATCHES[1],
    ]


@pytest.mark.asyncio
async def test_consumer_with_transformer(
    consumer_with_transformer: ConsumerWorker,
) -> None:
    consumer = consumer_with_transformer
    source = cast(StrSource, consumer.source)
    transformer = cast(DataTransformer, consumer.processor)
    target = cast(IntProcessor, transformer.target_processor)

    assert target.storage == []

    items_count = sum([len(batch) for batch in source.BATCHES])

    task = asyncio.create_task(consumer.run())

    while not task.done() and len(target.storage) < items_count:
        await asyncio.sleep(0)

    assert not task.done()
    task.cancel()

    assert target.storage == list(range(1, 11))
